@@ -91,14 +91,38 @@
 	+	April 2004 (johna	version 2.1.2
 	+	removed -Y in make install rule
 
-	+	August 2005 (johna)
+	+	August 2005 (johna)	version 2.1.3
 	+	added shm dump code for debugging broken libraries
 	+	added -z flag to invoke shm dump code
 
-	+	February 2006 (johna)
+	+	February 2006 (johna)	version 2.1.4
 	+	added code to handle new pedantic version of gcc
 	+	made signed/unsigned char assignments explicit
 
+	+	February 21, 2007 - djonas	version 2.1.5
+	+	updated noip2.c: added SkipHeaders() instead of the magic 6 line pass
+	+	Changed to ip1.dynupdate.no-ip.com for ip retrieval
+
+	+	August 2007 (johna)	version 2.1.6
+	+	added fclose() for stdin, stdout & stderr to child
+	+	made Force_Update work on 30 day intervals
+
+	+	August 2007 (johna)	version 2.1.7
+	+	fixed bug introduced in 2.1.6 where errors from multiple
+	+	instances were not diplayed due to stderr being closed
+	+	added version number into shared mem and -S display
+
+	+	November 2008 (johna)	still version 2.1.8
+	+       reworked forced update code to use 'wget' and the
+        +       hostactivate.php script at www.no-ip.com
+	+	added check of returned IP address in get_our_visble_IP_addr
+	+ 	hardened GetNextLine to prevent possible buffer overflow
+
+	+	November 2008 (johna)  version 2.1.9
+	+	hardened force_update() to prevent possible buffer overflow
+	+	hardened autoconf() the same way
+	+	patch suggested by xenomuta@phreaker.net
+
 */			
 /////////////////////////////////////////////////////////////////////////////                                            
 
@@ -153,13 +177,14 @@
 #define DEBUG
 
 #define ENCRYPT			1
+#define FORCE_UPDATE		0
 
 #define MAX(x,y)		(((x)>(y))?(x):(y))
 
 #define READ_TIMEOUT		90
 #define WRITE_TIMEOUT		60
 #define CONNECT_TIMEOUT		60
-#define FORCE_INTERVAL		(1440 * 30)		// 30 days in muinutes
+#define FORCE_INTERVAL		(1440 * 30)	// 30 days in minutes
 
 #define IPLEN			16
 #define LINELEN 	        256
@@ -167,14 +192,16 @@
 
 
 #define NOIP_NAME		"dynupdate.no-ip.com"
+#define UPD_NAME		"www.no-ip.com"
+#define UPD_SCRIPT		"hostactive.php"
+#define NOIP_IP_SCRIPT	"ip1.dynupdate.no-ip.com"
 #define CLIENT_IP_PORT		8245
 
-#define VERSION			"2.1.3"
+#define VERSION			"2.1.9"
 #define USER_AGENT		"User-Agent: Linux-DUC/"VERSION
 #define SETTING_SCRIPT		"settings.php?"
 #define USTRNG			"username="
 #define PWDSTRNG		"&pass="
-#define GETIP_SCRIPT		"ip.php"
 #if ENCRYPT
   #define REQUEST		"requestL="
 #else
@@ -204,6 +231,7 @@
 #define NODNSGROUP		"@@NO_GROUP@@"
 #define HOST			1
 #define GROUP			2
+#define DOMAIN			3
 #ifndef PREFIX
   #define PREFIX		"/usr/local"
 #endif
@@ -218,6 +246,7 @@
 #define NOIP_KEY		0x50494f4e
 #define SHMEM_SIZE		(MAX_INSTANCE * sizeof(struct INSTANCE))
 #define DEFAULT_NAT_INTERVAL	30
+#define WGET_PGM		"/usr/bin/wget"
 
 #define SPACE			' '
 #define EQUALS			'='
@@ -255,6 +284,7 @@
 #define BADCONFIG4		-14
 #define BADCONFIG5		-15
 #define BADCONFIG6		-16
+#define BADACTIVATE		-20
 
 #define CMSG01	"Can't locate configuration file %s. (Try -c). Ending!\n" 
 #define CMSG02	"'%s' is a badly constructed configuration file. Ending!\n" 
@@ -298,6 +328,9 @@
 #define CMSG40	"Do you wish to run something at successful update?[N] (y/N)  "
 #define CMSG40a	"Please enter the script/program name  "
 #define CMSG41	"Exec path unreadable in '%s'. Ending!\n" 
+#define CMSG42  "Can't get our visible IP address from %s"
+#define CMSG43  "Invalid domain name '%s'"
+#define CMSG44  "Invalid host name '%s'"
 #define CMSG99	"\nThis client has expired.  Please go to http://www.no-ip.com/downloads.php"  
 #define CMSG99a	"to get our most recent version.\n"  
 
@@ -328,7 +361,7 @@ int	kill_proc		=	0;
 int	reqnum			=	0;
 int	prompt_for_executable	=	1;
 int	shm_dump_active		=	0;
-int	Force_Update		=	0;
+int	Force_Update		=	FORCE_INTERVAL;
 int	dummy			=	0;
 void	*shmaddr		=	NULL;
 char	*program		=	NULL;
@@ -404,7 +437,8 @@ struct NAMES {
 
 struct INSTANCE {
 	int	pid;
-	short	debug;
+	char	debug;
+	char	version;
 	short	interval;
 	char	Last_IP_Addr[IPLEN];
 	char	cfilename[LINELEN];
@@ -432,13 +466,17 @@ unsigned char EncodeTable[] =
 
 ///////////////////////////////////////////////////////////////////////////
 void	process_options(int argc, char *argv[]);
+void	update_handler(int signum);
 void	alarm_handler(int signum);
 void	sigchld_handler(int signum);
 void	exit_handler(int signum);
 void	wake_handler(int signum);
+char	put_version(char *ver);
+char	*get_version(unsigned char x);
 void	save_IP();
 void	dump_shm(struct shmid_ds *d);
 int	get_shm_info();
+int	validate_IP_addr(char *src, char *dest);
 void	get_our_visible_IPaddr(char *dest);
 void	getip(char *dest, char *device);
 int	run_as_background();
@@ -457,6 +495,8 @@ int	parse_config();
 ushort	chksum(char *buf, unsigned int size);
 void	dump_buffer(int x);
 int	dynamic_update();
+int	validate_name(int flag, char *name);
+int	force_update();
 int	handle_dynup_error(int error_code);
 int	handle_config_error(int err_code);
 void	*Malloc(int size);
@@ -466,6 +506,7 @@ int	yesno(char *fmt, ...);
 int	add_to_request(int kind, char *p);
 int	get_update_selection(int tgrp, int thst);
 int	GetNextLine(char *dest);
+void SkipHeaders();
 void	url_encode(char *in, char *out);
 void	get_credentials(char *l, char *p);
 void	get_device_name(char *d);
@@ -566,6 +607,8 @@ int main(int argc, char *argv[])
 	if (background) {
 	    sig.sa_handler = (void *)wake_handler;
 	    sigaction(SIGUSR1,&sig,NULL);
+	    sig.sa_handler = update_handler;
+	    sigaction(SIGUSR2,&sig,NULL);
 	    sig.sa_handler = exit_handler;
 	    sigaction(SIGINT,&sig,NULL);
 	    sigaction(SIGTERM,&sig,NULL);
@@ -595,7 +638,7 @@ int main(int argc, char *argv[])
 		shmdt(shmaddr);			// done with our shared memory
 		return FATALERR;
 	    }
-	    strcpy(my_instance->Last_IP_Addr, IPaddress);
+	    strncpy(my_instance->Last_IP_Addr, IPaddress, IPLEN);
 	}
 	save_IP();
 	free(new_config);
@@ -729,6 +772,11 @@ void process_options(int argc, char *argv[])
 	return;
 }
 ///////////////////////////////////////////////////////////////////////////
+void update_handler(int signum)	// entered on SIGUSR2
+{
+	Force_Update = -1;
+}
+///////////////////////////////////////////////////////////////////////////
 void alarm_handler(int signum)	// entered on SIGALRM
 {
 	timed_out = 1;
@@ -752,6 +800,32 @@ void wake_handler(int signum)	//entered on SIGUSR1
 	dummy++;
 }
 ///////////////////////////////////////////////////////////////////////////
+char put_version(char *ver)
+{
+	char *p, x;
+	
+	p = ver + 2;	// step over 2.
+	x  = (*p -'0') * 10;	// get major num
+	p += 2;
+	x += (*p -'0');		// get minor num
+	return x;
+}
+///////////////////////////////////////////////////////////////////////////
+char *get_version(unsigned char z)
+{
+	int x, y;
+	static char vers[10];
+
+	if (z > 15) {
+	    x = z / 10;
+	    y = z % 10;
+	    sprintf(vers, "2.%d.%d", x, y);
+	} else {
+	    strcpy(vers, "unknown");
+	}
+	return vers;
+}
+///////////////////////////////////////////////////////////////////////////
 void save_IP()
 {
 	// config_file is opened as root, read-only permission has no effect
@@ -845,17 +919,17 @@ void dump_shm(struct shmid_ds *d)
 {
 	if (shm_dump_active) {
 	    fprintf(stderr, "\nShared memory info:\n");
-	    fprintf(stderr, "\tsize of struct shmid_ds is %d\n", sizeof(struct shmid_ds));
+	    fprintf(stderr, "\tsize of struct shmid_ds is %d\n", (int)sizeof(struct shmid_ds));
 	    fprintf(stderr, "\tmode %o\n", d->shm_perm.mode & 0x1ff);
 	    fprintf(stderr, "\tuid %d, gid %d\n", 
 					d->shm_perm.uid, d->shm_perm.gid);
 	    fprintf(stderr, "\tcreated by uid %d, gid %d\n", 
 					d->shm_perm.cuid, d->shm_perm.cgid);
-	    fprintf(stderr, "\tsegment size = %d\n", d->shm_segsz);
+	    fprintf(stderr, "\tsegment size = %d\n", (int)d->shm_segsz);
 	    fprintf(stderr, "\tlast attach time = %s",   ctime(&d->shm_atime));
 	    fprintf(stderr, "\tlast detach time = %s",   ctime(&d->shm_dtime));
 	    fprintf(stderr, "\tcreation time = %s",   ctime(&d->shm_ctime));
-	    fprintf(stderr, "\tnumber of attached procs = %ld\n", d->shm_nattch);
+	    fprintf(stderr, "\tnumber of attached procs = %d\n", (int)d->shm_nattch);
 	    fprintf(stderr, "\tcreation process id = %d\n", d->shm_cpid);
 	    fprintf(stderr, "\tlast user process id = %d\n", d->shm_lpid);
 	}
@@ -912,7 +986,7 @@ int get_shm_info()
 			}
 		    }
 		    if (!background && *IPaddress) { // update running daemon
-			strcpy(is->Last_IP_Addr, IPaddress);
+			strncpy(is->Last_IP_Addr, IPaddress, IPLEN);
 			continue;
 		    }
 		    Msg("Configuration '%s' already in use", is->cfilename);
@@ -929,6 +1003,7 @@ int get_shm_info()
 	my_instance->pid = getpid();
 	my_instance->debug = debug;
 	my_instance->interval = interval;
+	my_instance->version = put_version(VERSION);
 	sprintf(my_instance->Last_IP_Addr, new_config->lastIP);
 	sprintf(my_instance->args, "%s", saved_args);
 	sprintf(my_instance->cfilename, "%s", config_filename);
@@ -939,6 +1014,7 @@ int run_as_background()
 {
 	int	x, delay;
 	char	*err_string;
+	static int startup = 1;
 
 	x = fork();
 	switch (x) {
@@ -953,6 +1029,10 @@ int run_as_background()
 		if (get_shm_info() == FATALERR)	
 		    return FATALERR;
 		log2syslog++;
+		if (log2syslog > 0)
+		    fclose(stderr);
+		fclose(stdout);
+		fclose(stdin);
 		syslog(LOG_INFO, "v%s daemon started%s\n", 
 			VERSION, (nat) ?  " with NAT enabled" : "");
 		while (background) {
@@ -961,20 +1041,36 @@ int run_as_background()
 			get_our_visible_IPaddr(IPaddress);
 		    else
 			getip(IPaddress, device);
-#ifdef DEBUG
-		    if (my_instance->debug) 
-		        Msg("! Last_IP_Addr = %s, IP = %s",my_instance->Last_IP_Addr, IPaddress);
-#endif
-		    if (*IPaddress && ((Force_Update == 0) ||
-			    strcmp(IPaddress, my_instance->Last_IP_Addr)))  {
-			if (dynamic_update() == SUCCESS) 
-			    strcpy(my_instance->Last_IP_Addr, IPaddress);
-			*IPaddress = 0;
-			if (connect_fail)
-			    delay += 300;
+		    if (startup) {
+			startup = 0;
+			my_instance->Last_IP_Addr[0] =  '0';// force  check
 		    }
-		    if (--Force_Update < 0)
-			Force_Update = FORCE_INTERVAL / my_instance->interval;
+#ifdef DEBUG
+		    if (my_instance->debug)  {
+		        Msg("! Last_IP_Addr = %s, IP = %s",my_instance->Last_IP_Addr, IPaddress);
+#if FORCE_UPDATE
+			Msg("Force_Update == %d\n", Force_Update);
+#endif
+		    }
+#endif
+		    if (*IPaddress) { 
+			if (strcmp(IPaddress, my_instance->Last_IP_Addr)) {
+			    if (dynamic_update() == SUCCESS) 
+				strncpy(my_instance->Last_IP_Addr, IPaddress,
+									IPLEN);
+			    if (connect_fail)
+				delay = MAX(300, delay); // wait at least 5 minutes more
+			    *IPaddress = 0;
+			}
+		    }
+#if FORCE_UPDATE
+		    if (Force_Update < 0) {
+			if (force_update() == SUCCESS)
+			    Force_Update = FORCE_INTERVAL;
+		    } else {	// one time quanta closer to update trigger
+			Force_Update -= my_instance->interval;
+		    }
+#endif
 		    if (background)	// signal may have reset this!
 			Sleep(delay);
 		}
@@ -1319,7 +1415,8 @@ void display_current_config()
 			is = &shared->instance[i];
 			if (is->pid == 0)
 			    continue;
-			Msg("\nProcess %d, started as %s", is->pid, is->args);
+			Msg("\nProcess %d, started as %s, (version %s)", 
+				is->pid, is->args, get_version(is->version));
 			Msg("Using configuration from %s", is->cfilename);
 			Msg("Last IP Address set %s", is->Last_IP_Addr);
 			get_one_config(is->cfilename);
@@ -1411,6 +1508,38 @@ int parse_config()
 	return SUCCESS;
 }
 ///////////////////////////////////////////////////////////////////////////
+int validate_IP_addr(char *src, char *dest)
+{
+	char	*p;
+	int	i, octet;
+
+	octet = atoi(src);
+	if ((octet > 255) || (octet < 0))
+	    return 0;				// bad octet
+	if ((p = strchr(src, '.')) != NULL) {
+	    octet = atoi(++p);
+	    if ((octet > 255) || (octet < 0))
+		return 0;			// bad octet
+	    if ((p = strchr(p, '.')) != NULL) {
+		octet = atoi(++p);
+		if ((octet > 255) || (octet < 0))
+		    return 0;			// bad octet
+		if ((p = strchr(p, '.')) != NULL) {
+		    octet = atoi(++p);
+		    if ((octet <= 255) && (octet >= 0)) {
+			for (i=0; i<3; i++)
+			    if (!isdigit(p[i]))
+				break;
+			p[i] = 0;			// NULL terminate
+			strncpy(dest, src, IPLEN);
+			return 1;			// valid IP address
+		    }
+		}
+	    }
+	}
+	return 0;				// not enough dots
+}
+///////////////////////////////////////////////////////////////////////////
 void get_our_visible_IPaddr(char *dest)
 {
 	int	x;
@@ -1423,13 +1552,14 @@ void get_our_visible_IPaddr(char *dest)
 	} else {
 	    // set up text for web page query
 	    sprintf(buffer, 
-		"GET http://%s/ip.php HTTP/1.0\r\n%s\r\n\r\n", NOIP_NAME,USER_AGENT);
+		"GET http://%s/ HTTP/1.0\r\n%s\r\n\r\n", NOIP_IP_SCRIPT,USER_AGENT);
 	    if ((x = converse_with_web_server()) < 0) {
 		handle_dynup_error(x);
 	    } else {
 		close(socket_fd);
 		if ((p = strrchr(buffer,'\n'))) {// get end of penultimate line
-		    strcpy(dest, ++p);  	// address on next line
+		    if (!validate_IP_addr(++p, dest))// address on next line
+			Msg(CMSG42, NOIP_IP_SCRIPT); 
 #ifdef DEBUG
 		    if (my_instance ? my_instance->debug : debug)
 			fprintf(stderr,"! Our NAT IP address is %s\n", dest);
@@ -1438,7 +1568,7 @@ void get_our_visible_IPaddr(char *dest)
 		return;				// NORMAL EXIT
 	    } 
 	}
-	Msg("Can't get our visible IP address from %s", NOIP_NAME); 
+	Msg(CMSG42, NOIP_IP_SCRIPT); 
 	return;
 }
 ///////////////////////////////////////////////////////////////////////////
@@ -1465,6 +1595,7 @@ int dynamic_update()
 	}    
 	// set up text for web page update
 	bdecode(request, buffer);
+	// IPaddress has already been validated as to length and contents
 	sprintf(tbuf, "&ip=%s", IPaddress);
 	strcat(buffer, tbuf);
 	// use latter half of big buffer
@@ -1482,8 +1613,7 @@ int dynamic_update()
 
 	// analyze results
 	offset = 0;
-	for (i=0; i<6; i++) 
-	    GetNextLine(tbuf);
+	SkipHeaders();
 	response = 0;
 	while (GetNextLine(tbuf)) {
 	    p = ourname = tbuf;	// ourname points at the host/group or error
@@ -1501,7 +1631,7 @@ int dynamic_update()
 		*p++ = 0;		// now p points at return code
 		i = atoi(p);
 		if (is_group) {
-		    sprintf(gname, "group[%s]", ourname);
+		    snprintf(gname, LINELEN-1, "group[%s]", ourname);
 		    ourname = gname;
 		}
 		response++;
@@ -1527,6 +1657,7 @@ int dynamic_update()
 		    Msg("Can't fork (%s) to run %s", strerror(errno), buffer);
 		    exit(1);
 		case 0:			// child
+		    // IPaddress has already been validated as to length and contents
 		    sprintf(ipbuf, "%s", IPaddress);
 		    p = strrchr(buffer, '/');
 		    if (p)
@@ -1544,7 +1675,6 @@ int dynamic_update()
 		    break;
 	    }
 	}
-	Force_Update = -1;
 	return retval;
 }
 /////////////////////////////////////////////////////////////////////////////
@@ -1634,6 +1764,9 @@ int handle_dynup_error(int error_code)
 	    case WRITEFAIL:
 		Msg("Write to %s failed (%s)", NOIP_NAME, err_string); 
 		break;
+	    case BADACTIVATE:
+		Msg("Can't activate host at %s", UPD_NAME); 
+		break;
 	    default:
 		Msg("Unknown error %d trying to set %s at %s.", 
 						error_code, ourname, NOIP_NAME);
@@ -1847,7 +1980,7 @@ int add_to_request(int kind, char *p)
 {
 	char	tbuf[LINELEN];
 	
-	sprintf(tbuf, "&%s[]=%s", (kind==HOST) ? "h" : "g", p);
+	snprintf(tbuf, LINELEN-1, "&%s[]=%s", (kind==HOST) ? "h" : "g", p);
 	strcat(buffer, tbuf);
 	reqnum++;
 	return strlen(buffer);
@@ -1922,28 +2055,171 @@ int get_update_selection(int tgrp, int thst)
 	return len;
 }
 //////////////////////////////////////////////////////////////////////////
+void SkipHeaders()
+{
+	char *p;
+
+	// global offset into buffer, set here and used only by GetNextLine !
+	offset = 0;
+
+	// position after first blank line 
+	p = buffer;
+	for(; *p; p++) {
+		if(strncmp(p, "\r\n\r\n", 4) == 0) {
+			offset += 4;
+			break;
+		}
+		offset++;
+	}
+	return;
+}
+//////////////////////////////////////////////////////////////////////////
 int GetNextLine(char *dest)
 {
         char    *p = &buffer[offset];
         char    *q = dest;
+	char	*z = &dest[LINELEN-1];
 
         while (*p && (*p <= ' ')) {     // despace & ignore blank lines
             p++;
             offset++;
         }
-        while (*p) {
-            *q++ = *p;
+        while ((*p) && (q < z)) {
+            *q++ = (*p) & 0x7f;		// ASCII charset for text
             offset++;
             if (*p == '\n')  {
                 *q = 0;
 //fprintf(stderr, "LINE = %s", dest);
-                return 1;
+                return 1;		// we have a line
             } else
 		p++;
         }
-	if (q > dest)
-	    return 1;
-        return 0;
+	if (q > dest) {			// newline not found
+	    if (q == z)
+		q--;			// backup for newline and null
+	    *q++ = '\n';		// add '\n' 
+	    *q = 0;			// and null
+	    return 1;			// we have a line
+	}
+        return 0;			// no line available
+}
+///////////////////////////////////////////////////////////////////////////
+int force_update()
+{
+	int	x, found, retcode;
+	FILE	*pfd;
+	char	*l, *p, *q, domain[LINELEN], host[LINELEN];
+        char    ourbuffer[BIGBUFLEN], line[LINELEN], encline[2 * LINELEN];
+	
+	retcode = SUCCESS;
+	bdecode(request, buffer);
+	l = strchr(buffer, '=') + 1;		// point at login name
+	p = strchr(l, '&');
+	*p = 0;                                 // terminate it
+	p += 6;					// step over null and 'pass='
+	q = strchr(p, '&');
+	*q = 0;					// terminate password
+        sprintf(line, "%s%s%s%s", USTRNG, l, PWDSTRNG, p);
+        bencode(line, encline);
+
+	if ((x = Connect(port_to_use)) != SUCCESS) {
+	    handle_dynup_error(x);
+	    return x;
+	}    
+	// get the name list
+        sprintf(buffer, "GET http://%s/%s%s%s HTTP/1.0\r\n%s\r\n\r\n",
+                NOIP_NAME, SETTING_SCRIPT, REQUEST, encline, USER_AGENT);
+        if ((x = converse_with_web_server()) <= 0) {
+            handle_dynup_error(x);
+            return x;
+        }
+
+	SkipHeaders();
+        while (GetNextLine(line)) {
+            p = line;
+            if (strncmp(p, "<domain", 7) == 0)  {
+		p = strstr(line, "name=") + 6;
+		if (p == NULL) {
+		    Msg(CMSG44, p);
+		    continue;
+		}
+		q = strchr(p, '"');
+		if (q == NULL) {
+		    Msg(CMSG44, p);
+		    continue;
+		}
+		*q = 0;
+		if (!validate_name(DOMAIN, p)) {
+		    Msg(CMSG43, p);
+		    continue;
+		}
+		snprintf(domain, LINELEN-1, "%s", p);
+		continue;
+	    }
+            if (strncmp(p, "<host", 5) == 0)  {
+		p = strstr(line, "name=") + 6;
+		if (p == NULL) {
+		    Msg(CMSG44, p);
+		    continue;
+		}
+		q = strchr(p, '"');
+		if (q == NULL) {
+		    Msg(CMSG44, p);
+		    continue;
+		}
+		*q = 0;
+		if (!validate_name(HOST, p)) {
+		    Msg(CMSG44, p);
+		    continue;
+		}
+		snprintf(host, LINELEN-1, "%s", p);
+		// set up web page update command using wget
+		sprintf(ourbuffer, 
+		    "%s -q -O - 'http://%s/%s?host=%s&domain=%s'", 
+			WGET_PGM, UPD_NAME, UPD_SCRIPT, host, domain);
+#ifdef DEBUG
+		if (my_instance->debug) 
+		    Msg("> %s\n", ourbuffer);	// display the command
+#endif
+		// send wget command
+		pfd = popen(ourbuffer, "r");
+		found = 0;
+		while (fgets(ourbuffer, BIGBUFLEN, pfd)) {
+#ifdef DEBUG
+		    if (my_instance->debug) 
+			Msg("> %s\n", ourbuffer);   // display the line
+#endif
+		    if (strstr(ourbuffer, "Update Successful!!")) {
+			found = 1;
+			syslog(LOG_INFO, "Activation of host %s.%s succeeded", 
+								host, domain);
+			break;
+		    }
+		}
+		pclose(pfd);
+		if (!found) 
+		    syslog(LOG_INFO, "Can't activate host %s.%s",host, domain);
+	    }
+        }
+	close(socket_fd);
+	return retcode;
+}
+///////////////////////////////////////////////////////////////////////////
+int validate_name(int flag, char *name)
+{
+	char	 *p;
+	int	x;
+
+	if (flag == DOMAIN)
+	    x = '.';			// add the dot for domains
+	else
+	    x = '-';			// just re-use the hyphen for hosts
+	for (p=name; *p; p++) {
+	    if (isalnum(*p) || (*p == '-') || (*p == x))
+		continue;
+	    return 0;			// invalid char
+	}
+	return 1;
 }
 ///////////////////////////////////////////////////////////////////////////
 void url_encode(char *p_in, char *p_out)
@@ -2132,9 +2408,9 @@ int get_all_device_names(char *devs)
 	for (pIfr=Ifc.ifc_req,i=0; i<num_ifreq; pIfr++,i++) {
 	    q = dq;     // add new name into list
 	    p = (unsigned char *)pIfr->ifr_name;
-	    if (strncmp("lo", p, 2) == 0)	// ignore loopbacks
+	    if (strncmp("lo", (char *)p, 2) == 0)	// ignore loopbacks
 		continue;
-	    if (strchr(p, ':') != NULL)		// ignore 'virtual' devices
+	    if (strchr((char *)p, ':') != NULL)		// ignore 'virtual' devices
 		continue;
 	    devnum++;
 	    while (*p) 
@@ -2185,7 +2461,7 @@ void autoconf()
 	int	x, xfd;
 	int	a, b, c;
         FILE    *fd;
-        char    *p, *q, line[LINELEN], encline[2 * LINELEN], *pos;
+        char    *p, line[LINELEN], encline[2 * LINELEN], *pos;
 	char    internal_ip[IPLEN];
 	char    external_ip[IPLEN];
         struct SETTINGS *s;
@@ -2225,17 +2501,17 @@ void autoconf()
 	    Msg(CMSG27);
 	    return;
 	}
-        sprintf(buffer, "GET http://%s/%s\r\n\r\n", NOIP_NAME, GETIP_SCRIPT);
+        sprintf(buffer, "GET http://%s/\r\n\r\n", NOIP_IP_SCRIPT);
         if ((x = converse_with_web_server()) <= 0) {
             handle_dynup_error(x);
 	    return;
 	}
         p = buffer;
         if ((*p >= '0') && (*p <= '9')) {	// extract IP address
-            q = external_ip;
-            while (((*p >= '0') && (*p <= '9')) || (*p == '.'))
-                *q++ = *p++;
-	    *q = 0;
+	    if (!validate_IP_addr(buffer, external_ip)) {
+		Msg(CMSG44, p);
+		exit(1);
+	    }
         }
 #ifdef DEBUG
 	if (debug)
@@ -2267,8 +2543,7 @@ void autoconf()
         groups->glink = NULL;
         groups->nlink = NULL;
 
-	for (x=0; x<6; x++)	// skip first 6 lines HTTP protocol stuff
-            GetNextLine(line);
+	SkipHeaders();
         while (GetNextLine(line)) {
             p = line;
 	    s = settings;
@@ -2441,4 +2716,4 @@ void Msg(char *fmt, ...)
 		fprintf(stderr, "%s\n", errmsg);
 }
 /////////////////////////////////////////////////////////////////////////////
-
+/////////////////////////////////////////////////////////////////////////////
